@@ -26,13 +26,16 @@
 #include <gtk/gtk.h>
 #include <gdk/gdkx.h>
 #include <glib/gi18n.h>
+#include <X11/XKBlib.h>
 
 #include "lang.h"
 #include <time.h>
 #include <stdlib.h>
 #include <sys/stat.h>
+#include <sys/wait.h>
 
 #include "lxcom.h"
+#include "greeter-utils.h"
 
 enum {
     COL_SESSION_NAME,
@@ -47,8 +50,10 @@ enum {
     N_LANG_COLS
 };
 
-#define VCONFIG_FILE "/var/lib/lxdm/lxdm.conf"
+#define VCONFIG_FILE 		"/var/lib/lxdm/lxdm.conf"
+#define XKB_SYMBOL_DIR		"/usr/share/X11/xkb/symbols.dir"
 
+static GtkBuilder* builder;
 static GKeyFile *config;
 static GKeyFile * var_config;
 static GtkWidget* win;
@@ -66,17 +71,15 @@ static GtkWidget *lang_menu;
 
 static char* user = NULL;
 static char* pass = NULL;
-static char* session_exec = NULL;
-static char* session_desktop_file = NULL;
 
 static char* ui_file = NULL;
 static char *ui_nobody = NULL;
 
-static GdkPixbuf *bg_img = NULL;
-static GdkColor bg_color = {0};
-
 static GIOChannel *greeter_io;
 
+static int auto_login;
+static char datetime_fmt[8]="%c";
+
 static void do_reboot(void)
 {
     printf("reboot\n");
@@ -87,9 +90,34 @@ static void do_shutdown(void)
     printf("shutdown\n");
 }
 
-static void on_screen_size_changed(GdkScreen* scr, GtkWindow* win)
+static char *get_session_lang(void)
 {
-    gtk_window_resize( win, gdk_screen_get_width(scr), gdk_screen_get_height(scr) );
+	GtkTreeModel* model;
+	GtkTreeIter it;
+	gchar *res;
+	if(!lang)
+		return g_strdup("");
+	
+	if(!gtk_combo_box_get_active_iter(GTK_COMBO_BOX(lang), &it))
+		return g_strdup("");
+	model = gtk_combo_box_get_model(GTK_COMBO_BOX(lang));
+	gtk_tree_model_get(model, &it, 1, &res, -1);
+	return res;
+}
+
+static char *get_session_exec(void)
+{
+	GtkTreeModel* model;
+	GtkTreeIter it;
+	gchar *res;
+	if(!lang)
+		return g_strdup("");
+	
+	if(!gtk_combo_box_get_active_iter(GTK_COMBO_BOX(sessions), &it))
+		return g_strdup("");
+	model = gtk_combo_box_get_model(GTK_COMBO_BOX(sessions));
+	gtk_tree_model_get(model, &it, 1, &res, -1);
+	return res;
 }
 
 static void on_entry_activate(GtkEntry* entry)
@@ -110,44 +138,13 @@ static void on_entry_activate(GtkEntry* entry)
 	}
 	else
 	{
-		GtkTreeIter it;
-		char *session_lang = "";
-
-		if( gtk_combo_box_get_active_iter(GTK_COMBO_BOX(sessions), &it) )
-		{
-			GtkTreeModel* model = gtk_combo_box_get_model( GTK_COMBO_BOX(sessions) );
-			gtk_tree_model_get(model, &it, 1, &session_exec, 2, &session_desktop_file, -1);
-		}
-		else
-		{
-			/* FIXME: fatal error */
-		}
-
+		char *session_exec=get_session_exec();
+		char *session_lang=get_session_lang();
+		
 		tmp = g_strdup( gtk_entry_get_text(entry) );
 		pass=g_base64_encode((guchar*)tmp,strlen(tmp)+1);
 		g_free(tmp);
 
-		if( lang && gtk_combo_box_get_active_iter(GTK_COMBO_BOX(lang), &it) )
-		{
-			GtkTreeModel* model = gtk_combo_box_get_model( GTK_COMBO_BOX(lang) );
-			gtk_tree_model_get(model, &it, 1, &session_lang, -1);
-			//FIXME: is session leaked?
-		}
-
-		tmp = g_key_file_get_string(var_config, "base", "last_session", NULL);
-		if( g_strcmp0(tmp, session_desktop_file) )
-		{
-			g_key_file_set_string(var_config, "base", "last_session", session_desktop_file);
-		}
-		g_free(tmp);
-
-		tmp = g_key_file_get_string(var_config, "base", "last_lang", NULL);
-		if( g_strcmp0(tmp, session_lang) )
-		{
-			g_key_file_set_string(var_config, "base", "last_lang", session_lang);
-		}
-		g_free(tmp);
-
 		printf("login user=%s pass=%s session=%s lang=%s\n",
 			user, pass, session_exec, session_lang);
 
@@ -156,6 +153,8 @@ static void on_entry_activate(GtkEntry* entry)
 		user = NULL;
 		g_free(pass);
 		pass = NULL;
+		g_free(session_lang);
+		g_free(session_exec);
 
 		gtk_widget_hide(prompt);
 		gtk_widget_hide( GTK_WIDGET(entry) );
@@ -236,7 +235,11 @@ static void load_sessions()
 
 	g_free(last);
 	gtk_combo_box_set_model( GTK_COMBO_BOX(sessions), GTK_TREE_MODEL(list) );
+#if GTK_CHECK_VERSION(2,24,0)
+	gtk_combo_box_set_entry_text_column(GTK_COMBO_BOX(sessions), 0);
+#else
 	gtk_combo_box_entry_set_text_column(GTK_COMBO_BOX_ENTRY(sessions), 0);
+#endif
 	if( active_it.stamp )
 		gtk_combo_box_set_active_iter(GTK_COMBO_BOX(sessions), &active_it);
 	else
@@ -278,6 +281,17 @@ static gint lang_cmpr(GtkTreeModel *list,GtkTreeIter *a,GtkTreeIter *b,gpointer
 	return ret;
 }
 
+static gint keyboard_cmpr(GtkTreeModel *list,GtkTreeIter *a,GtkTreeIter *b,gpointer user_data)
+{
+	gint ret;
+	gchar *as,*bs;
+	gtk_tree_model_get(list,a,0,&as,-1);
+	gtk_tree_model_get(list,b,0,&bs,-1);
+	ret=strcmp(as,bs);
+	g_free(as);g_free(bs);
+	return ret;
+}
+
 static void on_menu_lang_select(GtkMenuItem *item,gpointer user_data)
 {
 	GtkTreeIter iter;
@@ -355,7 +369,7 @@ static void on_lang_changed(GtkComboBox *widget)
 	GtkTreeIter it;
 	if( gtk_combo_box_get_active_iter(widget, &it) )
 	{
-		GtkListStore *list=(GtkListStore*)gtk_combo_box_get_model(GTK_COMBO_BOX(lang));
+		GtkListStore *list=(GtkListStore*)gtk_combo_box_get_model(GTK_COMBO_BOX(widget));
 		char *lang=NULL;
 		gtk_tree_model_get(GTK_TREE_MODEL(list), &it, 1, &lang, -1);
 		if(lang[0]=='~')
@@ -401,13 +415,227 @@ static void load_langs()
     }
     g_free(lang_str);
     gtk_combo_box_set_model( GTK_COMBO_BOX(lang), GTK_TREE_MODEL(list) );
+#if GTK_CHECK_VERSION(2,24,0)
+	gtk_combo_box_set_entry_text_column(GTK_COMBO_BOX(lang), 0);
+#else
     gtk_combo_box_entry_set_text_column(GTK_COMBO_BOX_ENTRY(lang), 0);
+#endif
     gtk_combo_box_set_active(GTK_COMBO_BOX(lang), active < 0 ? 0 : active);
     g_object_unref(list);
 
     g_signal_connect(G_OBJECT(lang),"changed",G_CALLBACK(on_lang_changed),NULL);
 }
 
+static void on_keyboard_changed(GtkComboBox *widget)
+{
+	GtkTreeIter it;
+	if( gtk_combo_box_get_active_iter(widget, &it) )
+	{
+		GtkListStore *list=(GtkListStore*)gtk_combo_box_get_model(GTK_COMBO_BOX(widget));
+		char *keyboard=NULL;
+		char *cmd;
+		int status;
+		gboolean res;
+		gtk_tree_model_get(GTK_TREE_MODEL(list), &it, 0, &keyboard, -1);
+		/* set the current xkb */
+		cmd=g_strdup_printf("setxkbmap %s",keyboard);
+		res=g_spawn_command_line_sync(cmd,NULL,NULL,&status,NULL);
+		printf("%s %d %d\n",cmd,res,WEXITSTATUS (status));
+		g_free(cmd);
+		g_free(keyboard);
+	}
+}
+
+static gchar *xkb_name_norm(gchar *s)
+{
+	if(!strcmp(s,"pc")) return NULL;
+	if(!strncmp(s,"pc(",3)) return NULL;
+	if(!strncmp(s,"inet(",5)) return NULL;
+	if(!strncmp(s,"group(",6)) return NULL;
+	if(!strncmp(s,"srvr_ctrl(",10)) return NULL;
+	if(g_str_has_suffix(s,"(basic)"))
+	{
+		*strstr(s,"(basic)")=0;
+	}
+	else if(!strcmp(s,"jp(106)"))		//TODO: is default jp jp(106)
+	{
+		s[2]=0;
+	}
+	return s;
+}
+
+static char *xkb_get_current(void)
+{
+	Display *dpy=gdk_x11_display_get_xdisplay(gdk_display_get_default());
+	XkbDescRec * xkb_desc;
+	char *symbol_string=NULL;
+	gchar **list;
+	int i;
+
+	if(!dpy) return NULL;
+	xkb_desc=XkbAllocKeyboard();
+	if (xkb_desc == NULL)
+		return NULL;
+	XkbGetControls(dpy, XkbAllControlsMask, xkb_desc);
+	XkbGetNames(dpy, XkbSymbolsNameMask | XkbGroupNamesMask, xkb_desc);
+	if ((xkb_desc->names == NULL) || (xkb_desc->ctrls == NULL) || (xkb_desc->names->groups == NULL))
+	{
+	}
+	else
+	{
+		if (xkb_desc->names->symbols != None)
+		{
+			symbol_string=XGetAtomName(dpy, xkb_desc->names->symbols);
+		}
+	}
+	XkbFreeKeyboard(xkb_desc, 0, True);
+	
+	if(!symbol_string)
+		return FALSE;
+	list=g_strsplit(symbol_string,"+",-1);
+	XFree(symbol_string);
+	if(!list) return NULL;
+	for(i=0;list[i]!=NULL;i++)
+	{
+		if(!xkb_name_norm(list[i])) continue;
+		symbol_string=g_strdup(list[i]);
+		break;
+	}
+	g_strfreev(list);
+
+	return symbol_string;
+}
+
+#if 0
+
+static gboolean load_keyboards(GtkWidget *w)
+{
+	GtkListStore* list;
+	char p1[16],p2[16],p3[64];
+	FILE *fp;
+	char *cur;
+	int ret;
+	int count,active;
+	GtkTreeIter active_iter;
+	GdkScreen *scr;
+	
+	scr=gdk_screen_get_default();
+	if(gdk_screen_get_width(scr)<1024)
+		return FALSE;
+
+	if(!w) return FALSE;
+	
+	cur=xkb_get_current();
+	if(!cur) return FALSE;
+	fp=fopen(XKB_SYMBOL_DIR,"r");
+	if(!fp)
+	{
+		g_print("open xkb symbol dir fail\n");
+		g_free(cur);
+		return FALSE;
+	}
+	list = gtk_list_store_new(1, G_TYPE_STRING);
+	gtk_tree_sortable_set_sort_column_id(GTK_TREE_SORTABLE(list),0,GTK_SORT_ASCENDING);
+    gtk_tree_sortable_set_sort_func(GTK_TREE_SORTABLE(list),0,keyboard_cmpr,NULL,NULL);
+	for(count=0,active=-1;(ret=fscanf(fp,"%16s %16s %64s\n",p1,p2,p3))==3;)
+	{
+		GtkTreeIter iter;
+		if(strchr(p2,'m') && !strchr(p2,'a')) continue;
+		if(!xkb_name_norm(p3)) continue;
+		gtk_list_store_append(list,&iter);
+		gtk_list_store_set(list,&iter,0,p3,-1);
+		if(!strcmp(cur,p3))
+		{
+			active=count;
+			active_iter=iter;
+		}
+		count++;
+	}
+	fclose(fp);
+	g_free(cur);
+	
+	if(count==0 || active==-1)
+	{
+		g_object_unref(list);
+		return FALSE;
+	}
+	
+	gtk_combo_box_set_model(GTK_COMBO_BOX(w), GTK_TREE_MODEL(list) );
+	gtk_combo_box_entry_set_text_column(GTK_COMBO_BOX_ENTRY(w), 0);
+	g_object_unref(G_OBJECT(list));
+    gtk_combo_box_set_active_iter(GTK_COMBO_BOX(w), &active_iter);
+	
+	g_signal_connect(G_OBJECT(w),"changed",G_CALLBACK(on_keyboard_changed),NULL);
+
+	return TRUE;
+}
+#else
+static gboolean load_keyboards(GtkWidget *w)
+{
+	GtkListStore* list;
+	char *cur;
+	int count,active;
+	GtkTreeIter active_iter;
+	GdkScreen *scr;
+	GDir *dir;
+	const gchar *filename;
+	
+	scr=gdk_screen_get_default();
+	if(gdk_screen_get_width(scr)<1024)
+		return FALSE;
+
+	if(!w) return FALSE;
+	
+	cur=xkb_get_current();
+	if(!cur) return FALSE;
+	dir=g_dir_open("/usr/share/X11/xkb/symbols",0,FALSE);
+	if(!dir)
+	{
+		g_print("open xkb symbol dir fail\n");
+		g_free(cur);
+		return FALSE;
+	}
+	list = gtk_list_store_new(1, G_TYPE_STRING);
+	gtk_tree_sortable_set_sort_column_id(GTK_TREE_SORTABLE(list),0,GTK_SORT_ASCENDING);
+    gtk_tree_sortable_set_sort_func(GTK_TREE_SORTABLE(list),0,keyboard_cmpr,NULL,NULL);
+	for(count=0,active=-1;(filename=g_dir_read_name(dir))!=NULL;)
+	{
+		GtkTreeIter iter;
+		if(strlen(filename)!=2)
+			continue;
+		gtk_list_store_append(list,&iter);
+		gtk_list_store_set(list,&iter,0,filename,-1);
+		if(!strcmp(cur,filename))
+		{
+			active=count;
+			active_iter=iter;
+		}
+		count++;
+	}
+	g_dir_close(dir);
+	g_free(cur);
+	
+	if(count==0 || active==-1)
+	{
+		g_object_unref(list);
+		return FALSE;
+	}
+	
+	gtk_combo_box_set_model(GTK_COMBO_BOX(w), GTK_TREE_MODEL(list) );
+#if GTK_CHECK_VERSION(2,24,0)
+	gtk_combo_box_set_entry_text_column(GTK_COMBO_BOX(w), 0);
+#else
+	gtk_combo_box_entry_set_text_column(GTK_COMBO_BOX_ENTRY(w), 0);
+#endif
+	g_object_unref(G_OBJECT(list));
+    gtk_combo_box_set_active_iter(GTK_COMBO_BOX(w), &active_iter);
+	
+	g_signal_connect(G_OBJECT(w),"changed",G_CALLBACK(on_keyboard_changed),NULL);
+
+	return TRUE;
+}
+#endif
+
 static void on_exit_clicked(GtkButton* exit_btn, gpointer user_data)
 {
     gtk_menu_popup( GTK_MENU(exit_menu), NULL, NULL, NULL, NULL,
@@ -430,34 +658,39 @@ static void load_exit()
     g_signal_connect(exit_btn, "clicked", G_CALLBACK(on_exit_clicked), NULL);
 }
 
+#if 0
 static gboolean on_expose(GtkWidget* widget, GdkEventExpose* evt, gpointer user_data)
 {
     cairo_t *cr;
+	GdkRectangle *r=&evt->area;
 
+#if GTK_CHECK_VERSION(2,18,0)
+    if(! gtk_widget_get_has_window(widget))
+#else
     if( !GTK_WIDGET_REALIZED(widget) )
+#endif
         return FALSE;
+#if GTK_CHECK_VERSION(2,14,0)
+    cr = gdk_cairo_create(gtk_widget_get_window(widget));
+#else
     cr = gdk_cairo_create(widget->window);
-    if( bg_img )
+#endif
+    if(bg_img )
     {
-        cairo_matrix_t matrix;
-        double x = 0, y = 0, sx, sy;
-        cairo_get_matrix(cr, &matrix);
-        sx = (double)gdk_screen_width() / (double)gdk_pixbuf_get_width(bg_img);
-        sy = (double)gdk_screen_height() / (double)gdk_pixbuf_get_height(bg_img);
-        cairo_scale(cr, sx, sy);
-        gdk_cairo_set_source_pixbuf(cr, bg_img, x, y);
+        gdk_cairo_set_source_pixbuf(cr, bg_img, 0, 0);
+        gdk_cairo_rectangle (cr,r);
         cairo_paint(cr);
-        cairo_set_matrix(cr, &matrix);
     }
     else
     {
         gdk_cairo_set_source_color(cr, &bg_color);
-        cairo_rectangle( cr, 0, 0, gdk_screen_width(), gdk_screen_height() );
+        gdk_cairo_rectangle (cr,r);
         cairo_fill(cr);
     }
     cairo_destroy(cr);
     return FALSE;
 }
+#endif
 
 static gboolean on_combobox_entry_button_release(GtkWidget* w, GdkEventButton* evt, GtkComboBox* combo)
 {
@@ -474,22 +707,51 @@ static void fix_combobox_entry(GtkWidget* combo)
 {
     GtkWidget* edit = gtk_bin_get_child(GTK_BIN(combo));
     gtk_editable_set_editable( (GtkEditable*)edit, FALSE );
+#if GTK_CHECK_VERSION(2,18,0)
+    gtk_widget_set_can_focus(edit, FALSE);
+#else
     GTK_WIDGET_UNSET_FLAGS(edit, GTK_CAN_FOCUS);
+#endif
     g_signal_connect(edit, "button-release-event", G_CALLBACK(on_combobox_entry_button_release), combo);
 }
 
-static void on_evt_box_expose(GtkWidget* widget, GdkEventExpose* evt, gpointer user_data)
+#if GTK_CHECK_VERSION(3,0,0)
+static gboolean on_evt_box_draw(GtkWidget *widget,cairo_t *cr,gpointer user_data)
 {
+	GtkStyleContext *style=gtk_widget_get_style_context(widget);
+	gtk_render_background(style,cr,0,0,
+			gtk_widget_get_allocated_width(widget),
+			gtk_widget_get_allocated_height(widget));
+    return FALSE;
+}
+#else
+static gboolean on_evt_box_expose(GtkWidget *widget, GdkEventExpose* evt, gpointer user_data)
+{
+#if GTK_CHECK_VERSION(2,18,0)
+    if (gtk_widget_is_drawable(widget))
+#else
     if (GTK_WIDGET_DRAWABLE (widget))
+#endif
     {
         GtkWidgetClass* klass = (GtkWidgetClass*)G_OBJECT_GET_CLASS(widget);
-        gtk_paint_flat_box (widget->style, widget->window,
+        gtk_paint_flat_box (gtk_widget_get_style(widget),
+#if GTK_CHECK_VERSION(2,14,0)
+                            gtk_widget_get_window(widget),
+#else
+                            widget->window,
+#endif
+#if GTK_CHECK_VERSION(2,18,0)
+                gtk_widget_get_state(widget), GTK_SHADOW_NONE,
+#else
                 widget->state, GTK_SHADOW_NONE,
+#endif
                 &evt->area, widget, "eventbox",
                 0, 0, -1, -1);
         klass->expose_event (widget, evt);
     }
+    return FALSE;
 }
+#endif
 
 static gboolean on_timeout(GtkLabel* label)
 {
@@ -498,11 +760,40 @@ static gboolean on_timeout(GtkLabel* label)
     struct tm* tmbuf;
     time(&t);
     tmbuf = localtime(&t);
-    strftime(buf, 128, "%c", tmbuf);
+    strftime(buf, 128, datetime_fmt, tmbuf);
     gtk_label_set_text(label, buf);
     return TRUE;
 }
 
+static gboolean autologin_timeout(gpointer data)
+{
+	char *user=g_key_file_get_string(config,"base","autologin",NULL);
+	if(auto_login && user && user[0])
+	{
+		char *session_exec=get_session_exec();
+		char *session_lang=get_session_lang();
+
+		printf("autologin session=%s lang=%s\n",
+				session_exec, session_lang);
+
+		g_free(session_lang);
+		g_free(session_exec);
+
+		gtk_widget_hide(prompt);
+		gtk_widget_hide( GTK_WIDGET(login_entry) );
+	}
+	g_free(user);
+	return FALSE;
+}
+
+static gboolean is_autologin_user(const char *name)
+{
+	char *autologin=g_key_file_get_string(config,"base","autologin",NULL);
+	if(!autologin)
+		return FALSE;
+	return strcmp(name,autologin)?FALSE:TRUE;
+}
+
 static void on_user_select(GtkIconView *iconview)
 {
 	GList *list=gtk_icon_view_get_selected_items(iconview);
@@ -517,6 +808,23 @@ static void on_user_select(GtkIconView *iconview)
 	gtk_widget_hide(user_list);
 	if(name && name[0])
 	{
+		if(auto_login && is_autologin_user(name))
+		{
+			g_free(name);
+
+			char *session_exec=get_session_exec();
+			char *session_lang=get_session_lang();
+
+			printf("autologin session=%s lang=%s\n",
+					session_exec, session_lang);
+
+			g_free(session_lang);
+			g_free(session_exec);
+
+			gtk_widget_hide(prompt);
+			gtk_widget_hide( GTK_WIDGET(login_entry) );
+			return;
+		}
 		gtk_entry_set_text(GTK_ENTRY(login_entry),name);
 		g_free(name);
 		on_entry_activate(GTK_ENTRY(login_entry));
@@ -540,6 +848,12 @@ static void on_user_select(GtkIconView *iconview)
 		gtk_label_set_text( GTK_LABEL(prompt), _("User:") );
 		gtk_widget_show(prompt);
 	}
+	auto_login=0;
+}
+
+static void on_user_click(GtkIconView *iconview)
+{
+	on_user_select(iconview);
 }
 
 static gboolean load_user_list(GtkWidget *widget)
@@ -547,6 +861,7 @@ static gboolean load_user_list(GtkWidget *widget)
 	GtkListStore *model;
 	GtkTreeIter iter;
 	GKeyFile *kf;
+	GtkTreePath *path;	
 	char *res=NULL;
 	char **users;
 	gsize count;
@@ -570,12 +885,17 @@ static gboolean load_user_list(GtkWidget *widget)
 	gtk_icon_view_set_selection_mode(GTK_ICON_VIEW(widget),GTK_SELECTION_SINGLE);
 	gtk_icon_view_set_pixbuf_column(GTK_ICON_VIEW(widget),0);
 	gtk_icon_view_set_markup_column(GTK_ICON_VIEW(widget),1);
+#if GTK_CHECK_VERSION(2,22,0)
+	gtk_icon_view_set_item_orientation(GTK_ICON_VIEW(widget),GTK_ORIENTATION_HORIZONTAL);
+#else
 	gtk_icon_view_set_orientation(GTK_ICON_VIEW(widget),GTK_ORIENTATION_HORIZONTAL);
+#endif
 	model=gtk_list_store_new(5,GDK_TYPE_PIXBUF,G_TYPE_STRING,
 			G_TYPE_STRING,G_TYPE_STRING,G_TYPE_BOOLEAN);
 	gtk_icon_view_set_model(GTK_ICON_VIEW(widget),GTK_TREE_MODEL(model));
 	g_signal_connect(G_OBJECT(widget),"item-activated",G_CALLBACK(on_user_select),NULL);
-	g_signal_connect(G_OBJECT(widget),"selection-changed",G_CALLBACK(on_user_select),NULL);
+	//g_signal_connect(G_OBJECT(widget),"selection-changed",G_CALLBACK(on_user_select),NULL);
+	g_signal_connect(G_OBJECT(widget),"button-release-event",G_CALLBACK(on_user_click),NULL);
 	
 	users=g_key_file_get_groups(kf,&count);
 	if(!users || count<=0)
@@ -603,7 +923,11 @@ static gboolean load_user_list(GtkWidget *widget)
 		if(!face)
 		{
 			/* TODO: load some default face */
-			face=gdk_pixbuf_new_from_file_at_scale(ui_nobody,48,48,TRUE,NULL);
+			if(ui_nobody)
+				face=gdk_pixbuf_new_from_file_at_scale(ui_nobody,48,48,TRUE,NULL);
+			if(!face)
+				face=gtk_icon_theme_load_icon(gtk_icon_theme_get_default(),
+						"avatar-default", 48,GTK_ICON_LOOKUP_FORCE_SIZE,NULL);
 		}
 		display=g_strdup_printf("<span font_size=\"x-large\">%s</span>%s%s%s%s",
 			gecos?gecos:users[i],
@@ -624,46 +948,88 @@ static gboolean load_user_list(GtkWidget *widget)
 	// add "More ..."
 	gtk_list_store_append(model,&iter);
 	gtk_list_store_set(model,&iter,1,_("More ..."),2,"",3,"",4,FALSE,-1);
+	
+	path=gtk_tree_path_new_from_string("0");
+	gtk_icon_view_select_path(GTK_ICON_VIEW(widget),path);
+	gtk_widget_grab_focus(widget);
+	gtk_icon_view_set_cursor(GTK_ICON_VIEW(widget),path,NULL,FALSE);
+	gtk_tree_path_free(path);
 	return TRUE;
 }
 
+static void on_screen_size_changed(GdkScreen *screen,GtkWidget *win)
+{
+	GdkRectangle rc;
+	GdkWindow *window;
+#if GTK_CHECK_VERSION(3,0,0)
+	window=gtk_widget_get_window(win);
+#else
+	window=win->window;
+#endif
+	ui_get_geometry(window,&rc);
+	if(rc.width<1024 && g_key_file_get_integer(config, "display", "keyboard", 0)==1)
+	{
+		GtkWidget *w;
+		w=(GtkWidget*)gtk_builder_get_object(builder, "keyboard");
+		gtk_widget_hide(w);
+		w=(GtkWidget*)gtk_builder_get_object(builder, "label_keyboard");
+		gtk_widget_hide(w);
+	}
+	
+	gtk_window_move(GTK_WINDOW(win),rc.x,rc.y);
+	gtk_window_resize(GTK_WINDOW(win),rc.width,rc.height);
+	ui_set_bg(window,config);
+}
+
 static void create_win()
 {
-    GtkBuilder* builder;
-    GdkScreen* scr;
     GSList* objs, *l;
     GtkWidget* w;
+    GdkWindow *window;
+    gchar *temp;
+    GdkRectangle rc;
+    GdkScreen *scr;
+    
+    temp=g_key_file_get_string(config,"display","datetime",NULL);
+    if(temp && temp[0]=='%' && strlen(temp)<=3)
+		strcpy(datetime_fmt,temp);
+	g_free(temp);
 
     builder = gtk_builder_new();
     gtk_builder_add_from_file(builder, ui_file ? ui_file : LXDM_DATA_DIR "/lxdm.glade", NULL);
     win = (GtkWidget*)gtk_builder_get_object(builder, "lxdm");
+#if GTK_CHECK_VERSION(3,0,0)
+    gtk_window_set_has_resize_grip(GTK_WINDOW(win),FALSE);
+#endif
+    gtk_widget_realize(win);
+#if GTK_CHECK_VERSION(3,0,0)
+	window=gtk_widget_get_window(win);
+#else
+	window=win->window;
+#endif
 
-    /* set widget names according to their object id in GtkBuilder xml */
-    objs = gtk_builder_get_objects(builder);
-    for( l = objs; l; l = l->next )
-    {
-        GtkWidget* widget = (GtkWidget*)l->data;
-        gtk_widget_set_name( widget, gtk_buildable_get_name( (GtkBuildable*)widget ) );
-        char* path;
-        gtk_widget_path(widget, NULL, &path, NULL);
-    }
-    g_slist_free(objs);
+	/* set widget names according to their object id in GtkBuilder xml */
+	objs = gtk_builder_get_objects(builder);
+	for( l = objs; l; l = l->next )
+	{
+		char* path;
+		GtkWidget* widget = (GtkWidget*)l->data;
+		gtk_widget_set_name( widget, gtk_buildable_get_name( (GtkBuildable*)widget ) );
+		gtk_widget_path(widget, NULL, &path, NULL);
+	}
+	g_slist_free(objs);
 
-    if( bg_img ) /* only paint our own background if custom background image is set. */
+    if(g_key_file_has_key(config,"display","bg",NULL ))
     {
-        GTK_WIDGET_SET_FLAGS(win, GTK_APP_PAINTABLE);
-        g_signal_connect(win, "expose-event", G_CALLBACK(on_expose), NULL);
+#if GTK_CHECK_VERSION(3,0,0)
+		GtkStyleContext *style=gtk_widget_get_style_context(win);
+		gtk_style_context_remove_class(style,GTK_STYLE_PROPERTY_BACKGROUND_IMAGE);
+#endif
+        gtk_widget_set_app_paintable(win, TRUE);
+        
     } /* otherwise, let gtk theme paint it. */
 
-    scr = gtk_widget_get_screen(win);
-    g_signal_connect(scr, "size-changed", G_CALLBACK(on_screen_size_changed), win);
-    
     user_list=(GtkWidget*)gtk_builder_get_object(builder,"user_list");
-    if(user_list)
-    {
-		//TODO: load user list, connect select_changed,activate signal
-		// Use GtkIconView may enough
-	}
 
     prompt = (GtkWidget*)gtk_builder_get_object(builder, "prompt");
     login_entry = (GtkWidget*)gtk_builder_get_object(builder, "login_entry");
@@ -679,15 +1045,23 @@ static void create_win()
     if( g_key_file_get_integer(config, "display", "bottom_pane", 0) )
     {
         /* hacks to let GtkEventBox paintable with gtk pixmap engine. */
+#if GTK_CHECK_VERSION(2,18,0)
+        if(gtk_widget_get_app_paintable(w))
+#else
         if(GTK_WIDGET_APP_PAINTABLE(w))
-            g_signal_connect(w, "expose-event", G_CALLBACK(on_evt_box_expose), NULL);
+#endif
+
+#if GTK_CHECK_VERSION(3,0,0)
+		g_signal_connect(w,"draw",G_CALLBACK(on_evt_box_draw),NULL);
+#else
+		g_signal_connect(w, "expose-event", G_CALLBACK(on_evt_box_expose), NULL);
+#endif
     }
     else
         gtk_event_box_set_visible_window(GTK_EVENT_BOX(w), FALSE);
 
     if( g_key_file_get_integer(config, "display", "lang", 0) == 0 )
     {
-        GtkWidget *w;
         w = (GtkWidget*)gtk_builder_get_object(builder, "lang_box");
         if( w )
             gtk_widget_hide(w);
@@ -699,6 +1073,18 @@ static void create_win()
         fix_combobox_entry(lang);
         load_langs();
     }
+    
+    if(g_key_file_get_integer(config, "display", "keyboard", 0)==1)
+    {
+		w=(GtkWidget*)gtk_builder_get_object(builder, "keyboard");
+		if((load_keyboards(w))!=FALSE)
+		{
+			fix_combobox_entry(w);
+			gtk_widget_show(w);
+			w=(GtkWidget*)gtk_builder_get_object(builder, "label_keyboard");
+			if(w) gtk_widget_show(w);
+		}
+	}
 
 	if( (w = (GtkWidget*)gtk_builder_get_object(builder, "time"))!=NULL )
 	{
@@ -711,12 +1097,11 @@ static void create_win()
 	exit_btn = (GtkWidget*)gtk_builder_get_object(builder, "exit");
 	load_exit();
 
-	g_object_unref(builder);
+	ui_get_geometry(window,&rc);
+	gtk_window_move(GTK_WINDOW(win),rc.x,rc.y);
+	gtk_window_set_default_size(GTK_WINDOW(win),rc.width,rc.height);
+	ui_set_bg(window,config);
 
-	gtk_window_set_default_size( GTK_WINDOW(win), gdk_screen_get_width(scr), gdk_screen_get_height(scr) );
-	gtk_window_present( GTK_WINDOW(win) );
-	gtk_widget_realize(login_entry);
-    
 	if(user_list && !g_key_file_get_integer(config,"userlist","disable",NULL) && 
 			load_user_list(user_list))
 	{
@@ -729,57 +1114,17 @@ static void create_win()
 			gtk_widget_hide(user_list);
 			user_list=NULL;
 		}
-		gtk_widget_grab_focus(login_entry);
 	}
-}
-
-int set_background(void)
-{
-    char *bg;
-    char *style;
-    GdkWindow* root = gdk_get_default_root_window();
-    GdkCursor* cursor = gdk_cursor_new(GDK_LEFT_PTR);
-
-    gdk_window_set_cursor(root, cursor);
-
-    bg = g_key_file_get_string(config, "display", "bg", 0);
-    if( !bg )
-        bg = g_strdup("#222E45");
-    style = g_key_file_get_string(config, "display", "bg_style", 0);
-
-    if( bg )
-    {
-        if( bg[0] != '#' )
-        {
-            /* default the bg stretch */
-            if(!style || strcmp(style, "stretch") == 0 )
-            {
-                GdkPixbuf *tmp=gdk_pixbuf_new_from_file(bg,0);
-                if(tmp)
-                {
-                    bg_img=gdk_pixbuf_scale_simple(tmp,
-                                                    gdk_screen_width(),
-                                                    gdk_screen_height(),
-                                                    GDK_INTERP_HYPER);
-                    g_object_unref(tmp);
-                }
-            }
-            else
-            {
-                bg_img = gdk_pixbuf_new_from_file(bg, 0);
-            }
-            if( !bg_img )
-            {
-                g_free(bg);
-                bg = g_strdup("#222E45");
-            }
-        }
-        if( bg[0] == '#' )
-            gdk_color_parse(bg, &bg_color);
-    }
-    g_free(bg);
-    g_free(style);
-    return 0;
+	
+	ui_add_cursor();
+	ui_set_cursor(gtk_widget_get_window(win),GDK_LEFT_PTR);
+	gtk_widget_show(win);
+	ui_set_focus(window);
+	if(!user_list)
+		gtk_widget_grab_focus(login_entry);
+		
+	scr = gtk_widget_get_screen(win);
+	g_signal_connect(scr, "size-changed", G_CALLBACK(on_screen_size_changed), win);
 }
 
 static gboolean on_lxdm_command(GIOChannel *source, GIOCondition condition, gpointer data)
@@ -812,8 +1157,8 @@ static gboolean on_lxdm_command(GIOChannel *source, GIOCondition condition, gpoi
 		if(user_list)
 		{
 			gtk_widget_hide(login_entry);
-			gtk_icon_view_unselect_all(GTK_ICON_VIEW(user_list));
 			gtk_widget_show(user_list);
+			gtk_widget_grab_focus(user_list);
 		}
 		else
 		{
@@ -831,109 +1176,104 @@ void listen_stdin(void)
     g_io_add_watch(greeter_io, G_IO_IN, on_lxdm_command, NULL);
 }
 
-void set_root_background(void)
-{
-    GdkWindow *root = gdk_get_default_root_window();
-
-    /* set background */
-    if( !bg_img )
-    {
-        GdkColormap *map = (GdkColormap*)gdk_window_get_colormap(root);
-        gdk_color_alloc(map, &bg_color);
-        gdk_window_set_background(root, &bg_color);
-    }
-    else
-    {
-        GdkPixmap *pix = NULL;
-        gdk_pixbuf_render_pixmap_and_mask(bg_img, &pix, NULL, 0);
-        /* call x directly, because gdk will ref the pixmap */
-        //gdk_window_set_back_pixmap(root,pix,FALSE);
-        XSetWindowBackgroundPixmap( GDK_WINDOW_XDISPLAY(root),
-                                   GDK_WINDOW_XID(root), GDK_PIXMAP_XID(pix) );
-        g_object_unref(pix);
-    }
-    gdk_window_clear(root);
-}
-
 static void apply_theme(const char* theme_name)
 {
-    char* theme_dir = g_build_filename(LXDM_DATA_DIR "/themes", theme_name, NULL);
-    char* rc = g_build_filename(theme_dir, "gtkrc", NULL);
+	char* theme_dir = g_build_filename(LXDM_DATA_DIR "/themes", theme_name, NULL);
+	char *rc;
 
-    ui_file = g_build_filename(theme_dir, "greeter.ui", NULL);
+#if GTK_CHECK_VERSION(3,0,0)
+	ui_file = g_build_filename(theme_dir, "greeter-gtk3.ui", NULL);
+	rc=g_build_filename(theme_dir, "gtk.css", NULL);
+	GtkCssProvider *css=gtk_css_provider_new();
+	gtk_css_provider_load_from_path(css,rc,NULL);
+	g_free(rc);
+	gtk_style_context_add_provider_for_screen(gdk_screen_get_default(),
+			GTK_STYLE_PROVIDER(css),
+			GTK_STYLE_PROVIDER_PRIORITY_USER);
+#else
+	ui_file = g_build_filename(theme_dir, "greeter.ui", NULL);
+	rc = g_build_filename(theme_dir, "gtkrc", NULL);
+	if( g_file_test(rc, G_FILE_TEST_EXISTS) )
+	{
+		gtk_rc_parse(rc);
+	}
+	g_free(rc);
+#endif	
 
-    if( g_file_test(rc, G_FILE_TEST_EXISTS) )
-    {
-        gtk_rc_parse(rc);
-    }
-    g_free(rc);
+	if( !g_file_test(ui_file, G_FILE_TEST_EXISTS) )
+	{
+		g_free(ui_file);
+		ui_file = NULL;
+	}
+	
+#if GTK_CHECK_VERSION(3,0,0)
 
-    if( !g_file_test(ui_file, G_FILE_TEST_EXISTS) )
-    {
-        g_free(ui_file);
-        ui_file = NULL;
-    }
-    
-    ui_nobody = g_build_filename(theme_dir, "nobody.png", NULL);
-    if( !g_file_test(ui_nobody, G_FILE_TEST_EXISTS) )
-    {
-        g_free(ui_nobody);
-        ui_nobody = NULL;
-    }
-    
-    g_free(theme_dir);
-    
-    
+#endif
+
+	ui_nobody = g_build_filename(theme_dir, "nobody.png", NULL);
+	if( !g_file_test(ui_nobody, G_FILE_TEST_EXISTS) )
+	{
+		g_free(ui_nobody);
+		ui_nobody = NULL;
+	}
+
+	g_free(theme_dir);
 }
 
 int main(int arc, char *arg[])
 {
     char* theme_name;
-    GtkSettings*p;
+    GtkSettings *settings;
+    int i;
 
     /* this will override LC_MESSAGES */
     unsetenv("LANGUAGE");
 
+#if !GTK_CHECK_VERSION(3,0,0)
     gtk_set_locale();
+#endif
     bindtextdomain("lxdm", "/usr/share/locale");
     textdomain("lxdm");
 
     config = g_key_file_new();
     g_key_file_load_from_file(config, CONFIG_FILE, G_KEY_FILE_KEEP_COMMENTS, NULL);
-
     var_config = g_key_file_new();
     g_key_file_set_list_separator(var_config, ' ');
     g_key_file_load_from_file(var_config,VCONFIG_FILE,G_KEY_FILE_KEEP_COMMENTS, NULL);
 
     gtk_init(&arc, &arg);
+    for(i=1;i<arc;i++)
+    {
+		if(!strcmp(arg[i],"--auto-login"))
+		{
+			auto_login=g_key_file_get_integer(config,"base","timeout",NULL);
+		}
+	}
 
-    p=gtk_settings_get_default();
-    if(p)
+    settings=gtk_settings_get_default();
+    if(settings)
     {
         setenv("GTK_IM_MODULE","gtk-im-context-simple",1);
-        gtk_settings_set_string_property(p,"gtk-im-module","gtk-im-context-simple",0);
-        gtk_settings_set_long_property(p,"gtk-show-input-method-menu",0,0);
+        gtk_settings_set_string_property(settings,"gtk-im-module","gtk-im-context-simple",0);
+        gtk_settings_set_long_property(settings,"gtk-show-input-method-menu",0,0);
     }
 
-    set_background();
-    set_root_background();
-
     /* set gtk+ theme */
-    theme_name = g_key_file_get_string(config, "display", "gtk_theme", NULL);
-    if( theme_name )
-    {
-        GtkSettings* settings = gtk_settings_get_default();
-        g_object_set(settings, "gtk-theme-name", theme_name, NULL);
-        g_free(theme_name);
-    }
+	theme_name = g_key_file_get_string(config, "display", "gtk_theme", NULL);
+	if(theme_name)
+	{
+		if(settings)
+			g_object_set(settings, "gtk-theme-name", theme_name, NULL);
+		g_free(theme_name);
+	}   
 
     /* load gtkrc-based themes */
     theme_name = g_key_file_get_string(config, "display", "theme", NULL);
-    if( theme_name ) /* theme is specified */
-    {
-        apply_theme(theme_name);
-        g_free(theme_name);
-    }
+	if( theme_name ) /* theme is specified */
+	{
+		apply_theme(theme_name);
+		g_free(theme_name);
+	}
 
     /* create the login window */
     create_win();
@@ -941,7 +1281,11 @@ int main(int arc, char *arg[])
     /* use line buffered stdout for inter-process-communcation of
      * single-line-commands */
     setvbuf(stdout, NULL, _IOLBF, 0 );
-
+    
+    if(auto_login)
+    {
+		g_timeout_add_seconds(auto_login,autologin_timeout,NULL);
+	}
     gtk_main();
 
     g_key_file_free(config);
