@@ -33,6 +33,7 @@
 #include <string.h>
 #include <stdarg.h>
 #include <ctype.h>
+#include <stdint.h>
 #include <unistd.h>
 #include <stdarg.h>
 #include <fcntl.h>
@@ -53,7 +54,9 @@
 
 #include <execinfo.h>
 
-#include <utmp.h>
+#ifdef HAVE_UTMPX_H
+#include <utmpx.h>
+#endif
 
 #if HAVE_LIBPAM
 #include <security/pam_appl.h>
@@ -86,7 +89,7 @@ typedef struct{
 	CkConnector *ckc;
 #endif
 #ifndef DISABLE_XAUTH
-	char mcookie[33];
+	char mcookie[16];
 #endif
 	char **env;
 }LXSession;
@@ -150,6 +153,7 @@ void stop_pid(int pid)
     while( waitpid(-1, 0, WNOHANG) > 0 ) ;
 }
 
+#if HAVE_LIBPAM
 static void close_pam_session(pam_handle_t *pamh)
 {
     int err;
@@ -159,6 +163,7 @@ static void close_pam_session(pam_handle_t *pamh)
     pam_end(pamh, err);
     pamh = NULL;
 }
+#endif
 
 static LXSession *lxsession_find_greeter(void)
 {
@@ -433,15 +438,42 @@ static char *lxsession_xserver_command(LXSession *s)
 	if(!p) p=g_strdup("/usr/bin/X");	
 	g_shell_parse_argv(p, &arc, &arg, 0);
 	g_free(p);
-	for(i=1;i<arc;i++)
+	for(i=1;i<arc;)
 	{
-		g_free(arg[i]);
-		arg[i]=0;
+		p = arg[i];
+		if(!strncmp(p, "vt", 2) && isdigit(p[2]) &&
+			( !p[3] || (isdigit(p[3]) && !p[4]) ) )
+		{
+			g_free(arg[i]);
+			arc--;memcpy(arg+i,arg+i+1,(arc-i)*sizeof(char*));
+		}
+		else if(!strcmp(p,"-background"))
+		{
+			g_free(arg[i]);
+			arc--;memcpy(arg+i,arg+i+1,(arc-i)*sizeof(char*));
+			if(i<arc && !strcmp(arg[i],"none"))
+			{
+				g_free(arg[i]);
+				arc--;memcpy(arg+i,arg+i+1,(arc-i)*sizeof(char*));
+			}
+		}
+		else if(!strcmp(p,"-nr"))
+		{
+			g_free(arg[i]);
+			arc--;memcpy(arg+i,arg+i+1,(arc-i)*sizeof(char*));
+		}
+		else
+		{
+			i++;
+		}
 	}
+printf("arc %d\n",arc);
 	arg = g_renew(char *, arg, arc + 10);
-	arc=1;
 	if(nr_tty)
-		arg[arc++] = g_strdup("-nr");
+	{
+		arg[arc++] = g_strdup("-background");
+		arg[arc++] = g_strdup("none");
+	}
 	arg[arc++] = g_strdup_printf(":%d",s->display);
 	if(s->tty>0)
 		arg[arc++] = g_strdup_printf("vt%02d", s->tty);
@@ -463,7 +495,11 @@ void lxdm_get_tty(void)
 	gboolean plymouth;
     
 	plymouth=plymouth_is_running();
-	if(plymouth) plymouth_prepare_transition();
+	if(plymouth)
+	{
+		g_message("found plymouth running\n");
+		plymouth_prepare_transition();
+	}
 
 	old_tty=get_active_vt();
 	if( !s ) s = g_strdup("/usr/bin/X");
@@ -478,7 +514,7 @@ void lxdm_get_tty(void)
 			def_tty = atoi(p + 2);
 			gotvtarg = 1;
 		}
-		else if(!strcmp(p,"-nr"))
+		else if(!strcmp(p,"-background") || !strcmp(p,"-nr"))
 		{
 			nr_tty=1;
 		}
@@ -612,46 +648,65 @@ static void replace_env(char** env, const char* name, const char* new_val)
 }
 
 #ifndef DISABLE_XAUTH
-void create_server_auth(LXSession *s)
+
+static inline void xauth_write_uint16(int fd,uint16_t data)
+{
+	uint8_t t;
+	t=data>>8;
+	write(fd,&t,1);
+	t=data&0xff;
+	write(fd,&t,1);
+}
+
+static inline void xauth_write_string(int fd,const char *s)
+{
+	size_t len=strlen(s);
+	xauth_write_uint16(fd,(uint16_t)len);
+	write(fd,s,len);
+}
+
+static void xauth_write_file(const char *file,char data[16])
+{
+	int fd;
+	
+	fd=open(file,O_CREAT|O_TRUNC|O_WRONLY,0600);
+	if(!fd==-1) return;
+	xauth_write_uint16(fd,252);		//FamilyLocalHost
+	xauth_write_string(fd,"");
+	xauth_write_string(fd,"");
+	xauth_write_string(fd,"MIT-MAGIC-COOKIE-1");
+	xauth_write_uint16(fd,16);
+	write(fd,data,16);
+	close(fd);
+}
+
+static void create_server_auth(LXSession *s)
 {
 	GRand *h;
 	int i;
 	char *authfile;
 
 	h = g_rand_new();
-	const char *digits = "0123456789abcdef";
-	int r,hex=0;
-	for( i = 0; i < 31; i++ )
+	for( i = 0; i < 16; i++ )
 	{
-		r = g_rand_int(h) % 16;
-		s->mcookie[i] = digits[r];
-		if( r > 9 )
-			hex++;
+		s->mcookie[i] = g_rand_int(h)&0xff;
 	}
-	if( (hex % 2) == 0 )
-		r = g_rand_int(h) % 10;
-	else
-        	r = g_rand_int(h) % 5 + 10;
-	s->mcookie[31] = digits[r];
-	s->mcookie[32] = 0;
 	g_rand_free(h);
 
 	authfile = g_strdup_printf("/var/run/lxdm/lxdm-:%d.auth",s->display);
 
 	setenv("XAUTHORITY",authfile,1);
 	remove(authfile);
-	char *tmp = g_strdup_printf("xauth -q -f %s add %s . %s",
-                          authfile, getenv("DISPLAY"), s->mcookie);
-	g_spawn_command_line_sync(tmp,NULL,NULL,NULL,NULL);
-	g_free(tmp);
+	xauth_write_file(authfile,s->mcookie);
 	g_free(authfile);
 }
 
-void create_client_auth(char *home,char **env)
+static void create_client_auth(char *home,char **env)
 {
 	LXSession *s;
 	char *authfile;
 	uid_t user;
+	char *path;
 	
 	if((user=getuid())== 0 ) /* root don't need it */
 		return;
@@ -660,12 +715,18 @@ void create_client_auth(char *home,char **env)
 	if(!s)
 		return;
 
-	authfile = g_strdup_printf("%s/.Xauthority", home);
+	path=g_key_file_get_string(config,"base","xauth_path",NULL);
+	if(path)
+	{
+		authfile = g_strdup_printf("%s/.Xauth%d", path,getuid());
+		g_free(path);
+	}
+	else
+	{
+		authfile = g_strdup_printf("%s/.Xauthority", home);
+	}
 	remove(authfile);
-	char *tmp = g_strdup_printf("xauth -q -f %s add %s . %s",
-                          authfile, getenv("DISPLAY"), s->mcookie);
-	g_spawn_command_line_async(tmp,NULL);
-	g_free(tmp);
+	xauth_write_file(authfile,s->mcookie);
 	replace_env(env,"XAUTHORITY=",authfile);
 	g_free(authfile);
 }
@@ -1097,21 +1158,25 @@ static void exit_cb(void)
 
 static int get_run_level(void)
 {
+#if defined(HAVE_UTMPX_H) && defined(RUN_LVL)
 	int res=0;
-	struct utmp *ut,tmp;
+	struct utmpx *ut,tmp;
 
-	setutent();
+	setutxent();
 	tmp.ut_type=RUN_LVL;
-	ut=getutid(&tmp);
+	ut=getutxid(&tmp);
 	if(!ut)
 	{
-		endutent();
+		endutxent();
 		return 5;
 	}
 	res=ut->ut_pid & 0xff;
-	endutent();
+	endutxent();
 	//g_message("runlevel %c\n",res);
 	return res;
+#else
+	return 5;
+#endif
 }
 
 static void on_session_stop(void *data,int pid, int status)
@@ -1174,7 +1239,27 @@ gboolean lxdm_get_session_info(char *session,char **pname,char **pexec)
 		}
 		else
 		{
-			name=g_strdup(session);
+			GKeyFile *f;
+			char *file_path;
+			gboolean loaded;
+
+			f = g_key_file_new();
+			char *desktop_name = g_strconcat(session, ".desktop", NULL);
+			file_path = g_build_filename(XSESSIONS_DIR, desktop_name, NULL);
+			loaded = g_key_file_load_from_file(f, file_path, G_KEY_FILE_NONE, NULL);
+			g_free(file_path);
+			g_free(desktop_name);
+
+			if ( loaded )
+			{
+				name = g_key_file_get_locale_string(f, "Desktop Entry", "Name", NULL, NULL);
+				exec = g_key_file_get_string(f, "Desktop Entry", "Exec", NULL);
+			}
+			else
+			{
+				name=g_strdup(session);
+			}
+			g_key_file_free(f);
 		}
 	}
 	if(name && !exec)
@@ -1331,17 +1416,25 @@ void lxdm_do_login(struct passwd *pw, char *session, char *lang, char *option)
 	s->user=pw->pw_uid;
 	if(option)
 		s->option=g_strdup(option);
+#if HAVE_LIBCK_CONNECTOR
 	if(s->ckc)
 	{
 		ck_connector_unref(s->ckc);
 		s->ckc=NULL;
 	}
+#endif
 #if HAVE_LIBPAM
 	setup_pam_session(s,pw,session_name);
 #endif
 #if HAVE_LIBCK_CONNECTOR
+#if HAVE_LIBPAM
+	if(!s->ckc && (!s->pamh || !pam_getenv(s->pamh,"XDG_SESSION_COOKIE")))
+#else
 	if(!s->ckc)
+#endif
+	{
 		s->ckc = ck_connector_new();
+	}
 	if( s->ckc != NULL )
 	{
 		DBusError error;
@@ -1432,6 +1525,11 @@ void lxdm_do_shutdown(void)
 	lxdm_quit_self(0);
 }
 
+static gboolean auto_login_future(void)
+{
+	return g_key_file_get_integer(config,"base","timeout",NULL)>=5;
+}
+
 int lxdm_do_auto_login(void)
 {
 	struct passwd *pw;
@@ -1531,6 +1629,27 @@ static gboolean strv_find(char **strv,const char *p)
 	return FALSE;
 }
 
+static char *lxdm_get_user_face(struct passwd *pw)
+{
+	GKeyFile *kf;
+	char *face;
+	char *file;
+	face=g_strdup_printf("%s/.face",pw->pw_dir);
+	if(!access(face,R_OK))
+		return face;
+	g_free(face);
+	kf=g_key_file_new();
+	file=g_build_filename ("/var/lib/AccountsService/users/", pw->pw_name, NULL);
+	g_key_file_load_from_file (kf, file, 0, NULL);
+	g_free(file);
+	face=g_key_file_get_string(kf,"User","Icon",NULL);
+	g_key_file_free(kf);
+	if(face && !access(face,R_OK))
+		return face;
+	g_free(face);
+	return NULL;
+}
+
 GKeyFile *lxdm_user_list(void)
 {
 	struct passwd *pw;
@@ -1576,8 +1695,8 @@ GKeyFile *lxdm_user_list(void)
 			g_key_file_set_string(kf,pw->pw_name,"gecos",pw->pw_gecos);
 		if(lxsession_find_user(pw->pw_uid))
 			g_key_file_set_boolean(kf,pw->pw_name,"login",TRUE);
-		face=g_strdup_printf("%s/.face",pw->pw_dir);
-		if(!access(face,R_OK))
+		face=lxdm_get_user_face(pw);
+		if(face)
 			g_key_file_set_string(kf,pw->pw_name,"face",face);
 		g_free(face);
 	}
@@ -1702,7 +1821,8 @@ int main(int arc, char *arg[])
 	set_signal();
 	lxdm_get_tty();
 
-	lxdm_do_auto_login();
+	if(!auto_login_future())
+		lxdm_do_auto_login();
 	if(!lxsession_get_active())
 		lxsession_greeter();
 
